function [optsm1,optsm2,other] = cv_mcatpp_cyc(x,basis,p)
% [optsm1,optsm2,other] = cv_mcatpp_cyc(x,basis,p)
%
% Cross-validation search of smoothing parameters for temporal MCA
%   with cyclic border condition
%  (Five-fold cross-validation is used)
%
% INPUT:
%   x: Observed time points (n x 1 cell).
%      Each x{i} is a vector containing the data from replication i.
%   basis: B-spline basis parameters. Struct with the following fields:
%       rng: Time range (1 x 2 vector).
%       or: Spline order (integer; 4 is cubic splines).
%       nk: Number of knots (integer). Knots will be equally spaced.
%   p: Number of model components (integer>=0).
%
% OUTPUT:
%   optsm1: Optimal smoothing parameter for the mean (scalar>=0)
%   optsm2: Optimal smoothing parameter for components (scalar>=0)
%            (Returns [] if p=0).
%   other: Additional output. Struct with the following fields:
%       optOF1: Optimal value of the objective function at optsm1.
%       optOF2: Optimal value of the objective function at optsm2.
%                (Returns [] if p=0).
%       smgrid: Grid of smoothing parameters used.
%       OF1grid: Objective function at smgrid for mean-only models.
%       OF2grid: Objective function at smgrid for p-component models.
%                 (Returns [] if p=0).
%
% Programs called: MCATPP_CYC, PRED_MCATPP
%
% Version: June 2018

% Cross-validation
itmax = 10;
smgrid = 10.^(-7:.5:-1)';
Ng = length(smgrid);
% Find optimal sm for mean
disp('Finding optimal sm1')
OF1grid = -Inf(Ng,1);
optOF1 = -Inf;
optsm1 = -Inf;
for i = 1:Ng
    disp(['Cross-validating for grid point ' num2str(i) ' of ' num2str(Ng)])
    logf = cv_5f(x,basis,0,smgrid(i),0,itmax);
    OF1grid(i) = mean(logf);
    if OF1grid(i)>optOF1
        optOF1 = OF1grid(i);
        optsm1 = smgrid(i);
    end
end
% Find optimal sm for components
disp(' ')
disp('Finding optimal sm2')
if p==0
    optOF2 = [];
    optsm2 = [];
    OF2grid = [];
else
    OF2grid = -Inf(Ng,1);
    optOF2 = -Inf;
    for i = 1:Ng
        disp(['Cross-validating for grid point ' num2str(i) ' of ' num2str(Ng)])
        logf = cv_5f(x,basis,p,optsm1,smgrid(i),itmax);
        OF2grid(i) = mean(logf);
        if OF2grid(i)>optOF2
            optOF2 = OF2grid(i);
            optsm2 = smgrid(i);
        end
    end
end
% Output
other = struct('optOF1',optOF1,'optOF2',optOF2,'smgrid',smgrid,...
    'OF1grid',OF1grid,'OF2grid',OF2grid);
end

%%%% -------- AUXILIARY FUNCTIONS

function logf = cv_5f(x,basis,p,sm1,sm2,itmax)
% Five-fold cross-validation for MCATPP
% Computes cross-validated log-densities
% Programs called: MCATPP_CYC, PRED_MCATPP
n = length(x);
logf = -Inf(n,1);
B = round(n/5);
for i = 1:5
    itest = ((i-1)*B+1):(i*B);
    if i==5
        itest = ((i-1)*B+1):n;
    end
    itrain = setdiff(1:n,itest);
    try
        [T,c0,C,s2] = evalc('mcatpp_cyc(x(itrain),basis,p,sm1,sm2,itmax)');
    catch ME1
        c0 = [];
        C = [];
        s2 = [];
    end
    if ~isempty(c0)
        [T,logf(itest)] = evalc('pred_mcatpp(x(itest),basis,c0,C,s2)');
    end
end
end
