function [c0,C,s2,u,logf] = mcatpp(x,basis,p,sm1,sm2,itmax)
% [c0,C,s2,u,logf] = mcatpp(x,basis,p,sm1,sm2,itmax)
%
% Multiplicative Component Analysis for Temporal Point Processes
%   (PCA of log-intensities)
%
% INPUT:
%   x: Observed time points (n x 1 cell).
%      Each x{i} is a vector containing the data from replication i.
%   basis: B-spline basis parameters. Struct with the following fields:
%       rng: Time range (1 x 2 vector).
%       or: Spline order (integer; 4 is cubic splines).
%       nk: Number of knots (integer). Knots will be equally spaced.
%   p: Number of model components (integer>=0).
%   sm1: Smoothing parameter for the mean (scalar>=0).
%   sm2: Smoothing parameter for the components (scalar>=0).
%        (All components have norm 1 but the mean does not, so different
%         sm's may be needed to attain the same degree of smoothness).
%   itmax: Maximum number of iterations (integer).
%
% OUTPUT:
%   c0: Mean basis coefficients (q x 1).
%   C: Component basis coefficients (q x p).
%   s2: Component variances (p x 1).
%   u: Individual component scores (n x p).
%   logf: Individual log-densities (n x 1).
%
% External calls: BSPL
%
% Version: June 2018

% Input check
c0 = [];
C = [];
s2 = [];
u = [];
logf = [];
if ~iscell(x)
    disp('Error: X must be cell array')
    return
else
    [mx,nx] = size(x);
    if (mx>1 && nx>1)
        disp('Error: X must be a one-dimensional cell array')
        return
    end
end
n = length(x);

% Data filtering (elimination of data outside RNG)
m = zeros(n,1);
a = basis.rng(1);
b = basis.rng(2);
for i = 1:n
    x{i}(x{i}<a) = [];
    x{i}(x{i}>b) = [];
    m(i) = length(x{i});
end
if any(m==0)
    disp('Warning: Some x{i}s have no data within basis range')
end
if any(m>=200)
    disp('Warning: Some x{i}s have more than 200 observations')
    disp('This may cause Inf values in the likelihood function')
    disp('This method is intended for relatively small x{i}s')
    disp('For large x{i}s you can just use kernel smoothing')
end

% Initialization
q = basis.or + basis.nk;
t = linspace(a,b,300);
dt = t(2)-t(1);
knt = linspace(a,b,basis.nk+2);
B0 = bspl(t,basis.or,knt,0);
J0 = (B0'*B0)*dt;
if basis.or>2
    B2 = bspl(t,basis.or,knt,2);
    J2 = (B2'*B2)*dt;
else
    J2 = zeros(q,q);
end
sumB = zeros(n,q);
for i = 1:n
    B_i = bspl(x{i},basis.or,knt,0);
    sumB(i,:) = sum(B_i,1);
end

% -----> Initial mean-only model
c0 = log(mean(m)/(b-a))*ones(q,1);
logf = complogf0(sumB,c0,dt,B0,m);
OF = mean(logf)-sm1*c0'*J2*c0;
disp('---> Computing mean')
disp(['Iteration: 0, Pen. loglik: ' num2str(OF)])
err = 1;
iter = 0;
while err>1e-3 && iter<itmax
    iter = iter + 1;
    c00 = c0;
    OF0 = OF;
    [gc,Hc] = derivc0(sumB,c0,dt,B0);
    gpll = gc-2*sm1*J2*c0;
    Hpll = Hc-2*sm1*J2;
    direction = Hpll\gpll;
    OF = -Inf;
    k = 0;
    while OF<=OF0 && k<6
        step = 0.7^k;
        c0 = c00-step*direction;
        logf = complogf0(sumB,c0,dt,B0,m);
        OF = mean(logf)-sm1*c0'*J2*c0;
        k = k+1;
    end
    % Stopping criterion
    if OF<=OF0 || ~all(isfinite(c0))
        disp('No further improvement in obj. func. is possible')
        c0 = c00;
        OF = OF0;
    end
    err = norm(c0-c00)/norm(c00);
    disp(['Iteration: ' num2str(iter) ', Pen. loglik: ' ...
            num2str(OF) ', Error: ' num2str(err)])
end

% -----> Sequential PC estimation
C = zeros(q,p);
s2 = zeros(p,1);
u = zeros(n,p);
u2 = zeros(n,p);
for ic = 1:p
    if ic==1
        P = eye(q);
    else
        P = null(C(:,1:ic-1)'*J0);
    end
    % Initial estimators    
    C(:,ic) = (P*P')*ones(q,1);
    C(:,ic) = C(:,ic)/sqrt(C(:,ic)'*J0*C(:,ic));
    if ic==1
        Ilmb0 = sum(exp(B0*c0))*dt;
        u(:,ic) = sqrt(b-a)*log(max(m,1)/Ilmb0);
        s2(ic) = var(u(:,ic));
    else
        s2(ic) = s2(ic-1)/2;
    end
    [u(:,1:ic),u2(:,1:ic),logf] = ...
        compeff(sumB,c0,C(:,1:ic),s2(1:ic),dt,B0,m,u(:,1:ic));
    OF = mean(logf)-sm1*c0'*J2*c0-sm2*sum(diag(C(:,1:ic)'*J2*C(:,1:ic)));
    disp(['---> Computing component ' num2str(ic)])
    disp(['Iteration: 0, Pen. loglik: ' num2str(OF)])
    err = 1;
    iter = 0;
    % Iterations
    while err>1e-3 && iter<itmax
        iter = iter + 1;
        % Update C
        c00 = C(:,ic);
        u00 = u;
        u200 = u2;
        OF0 = OF;
        [gc,Hc] = derivc(sumB,c0,C(:,1:ic),dt,B0,u(:,1:ic),u2(:,1:ic));
        gpll = gc-2*sm2*J2*C(:,ic);
        Hpll = Hc-2*sm2*J2;
        direction = P*((P'*Hpll*P)\(P'*gpll));
        OF = -Inf;
        k = 0;
        while OF<=OF0 && k<6
            step = 0.7^k;
            C(:,ic) = c00-step*direction;
            C(:,ic) = C(:,ic)/sqrt(C(:,ic)'*J0*C(:,ic));
            [u(:,1:ic),u2(:,1:ic),logf] = ...
                compeff(sumB,c0,C(:,1:ic),s2(1:ic),dt,B0,m,u00(:,1:ic));
            OF = mean(logf) - sm1*c0'*J2*c0 ...
                -sm2*sum(diag(C(:,1:ic)'*J2*C(:,1:ic)));
            k = k+1;
        end
        if OF<=OF0 || ~all(isfinite(C(:,ic)))
            disp('No further improvement in obj. func. is possible')
            C(:,ic) = c00;
            u = u00;
            u2 = u200;
        end
        % Update s2
        s2 = mean(u2,1)';
        % Stopping criterion
        err = norm(C(:,ic)-c00)/norm(c00);
        disp(['Iteration: ' num2str(iter) ', Pen. loglik: ' ...
            num2str(OF) ', Error: ' num2str(err)])
    end
end
end

%%%%%%%------ AUXILIARY FUNCTIONS

function logf = complogf0(sumB,c0,dt,B0,m)
%Computes log-densities for mean-only model
logf = -sum(exp(B0*c0))*dt + sumB*c0 - gammaln(m+1);
end

function [gc0,Hc0] = derivc0(sumB,c0,dt,B0)
% Derivatives of loglik/n w.r.t c0
q = size(B0,2);
Hc0 = -(B0'*((exp(B0*c0)*ones(1,q)).*B0))*dt;
gc0 = -B0'*exp(B0*c0)*dt + mean(sumB,1)';
end

function [u,u2,logf] = compeff(sumB,c0,C,s2,dt,B0,m,u_ini)
%Computes random effects and log-pdf using Laplace approximation
[n,p] = size(u_ini);
Phi = B0*C;
logf = zeros(n,1);
u = u_ini;
u2 = u_ini.^2;
for i = 1:n
    % Compute log(f(x))
    uL = u_ini(i,:);
    D_gi = zeros(1,p);
    H_gi = eye(p);
    for steps = 1:5
        uL = uL - D_gi/H_gi;
        lmbi = exp(B0*c0+B0*C*uL');
        D_gi = -lmbi'*Phi*dt + sumB(i,:)*C - uL./s2';
        H_gi = -Phi'*((lmbi*ones(1,p)).*Phi)*dt - diag(1./s2);
    end
    gi = -sum(lmbi)*dt + sumB(i,:)*(c0+C*uL') - gammaln(m(i)+1) ...
        -sum(uL.^2./(2*s2')) - .5*sum(log(2*pi*s2));
    logf(i) = gi + (p/2)*log(2*pi) - .5*logdet(-H_gi);
    S = (-H_gi)\eye(p);
    u(i,:) = uL;
    u2(i,:) = diag(S)' + uL.^2;
end
end

function [gc,Hc] = derivc(sumB,c0,C,dt,B0,u,u2)
% Derivatives of loglik/n w.r.t C(:,end)
% Uses ad-hoc approx of second derivatives and plug-in scores for integrals
[n,p] = size(u);
q = length(c0);
ng = size(B0,1);
lmb = exp(B0*c0*ones(1,n)+B0*C*u');
ulmb = (ones(ng,1)*u(:,p)').*lmb;
u2lmb = (ones(ng,1)*u2(:,p)').*lmb;
gc = (-B0'*mean(ulmb,2))*dt + (sumB'*u(:,p)/n);
Hc = -(B0'*((mean(u2lmb,2)*ones(1,q)).*B0))*dt;
end

function y = logdet(A)
% log(det(A)) for symmetric non-neg A
R = chol(A);
y = 2*sum(log(diag(R)));
end

