function [logf,u] = pred_mcatpp(x,basis,c0,C,s2)
% [logf,u] = pred_mcatpp(x,basis,c0,C,s2)
%
% Random-effect predictors for temporal Multiplicative Component models
%   (Typically used for cross-validation)
%
% INPUT:
%   x: Time points (typically "test" data) (n x 1 cell).
%       Each x{i} is a vector of observations for replication i.
%   basis: B-spline basis parameters. Struct with the following fields:
%       rng: Time range (1 x 2 vector).
%       or: Spline order (integer; 4 is cubic splines).
%       nk: Number of knots (integer). Knots will be equally spaced.
%   c0: Mean basis coefficients (q x 1).
%   C: Component basis coefficients (q x p). Use [] for mean-only model.
%   s2: Component variances (p x 1). Use [] for mean-only model.
%
% OUTPUT:
%   logf: Individual log-density values (n x 1).
%   u: Predicted component scores (n x p).
%
% External calls: BSPL
%
% Version: June 2018

% Input check
logf = [];
u = [];
if ~iscell(x)
    disp('Error: X must be cell array')
    return
else
    [mx,nx] = size(x);
    if (mx>1 && nx>1)
        disp('Error: X must be a one-dimensional cell array')
        return
    end
end
n = length(x);

% Data filtering (elimination of data outside RNG)
m = zeros(n,1);
a = basis.rng(1);
b = basis.rng(2);
for i = 1:n
    x{i}(x{i}<a) = [];
    x{i}(x{i}>b) = [];
    m(i) = length(x{i});
end
if any(m==0)
    disp('Warning: Some individuals have no data within basis range')
end

% Preliminary computations
t = linspace(a,b,300);
dt = t(2)-t(1);
knt = linspace(a,b,basis.nk+2);
B0 = bspl(t,basis.or,knt,0);
q = size(B0,2);
sumB = zeros(n,q);
for i = 1:n
    B_i = bspl(x{i},basis.or,knt,0);
    sumB(i,:) = sum(B_i,1);
end

% Main computations
if isempty(C)
    % Mean-only model
    logf = complogf0(sumB,c0,dt,B0,m);
    u = [];
else
    % At-least-one-component model
    p = size(C,2);
    u_ini = zeros(n,p);
    [u,u2,logf] = compeff(sumB,c0,C,s2,dt,B0,m,u_ini);
end
end

%%%%%%%------ AUXILIARY FUNCTIONS

function logf = complogf0(sumB,c0,dt,B0,m)
%Computes log-densities for mean-only model
logf = -sum(exp(B0*c0))*dt + sumB*c0 - gammaln(m+1);
end

function [u,u2,logf] = compeff(sumB,c0,C,s2,dt,B0,m,u_ini)
%Computes random effects and log-pdf using Laplace approximation
[n,p] = size(u_ini);
Phi = B0*C;
logf = zeros(n,1);
u = u_ini;
u2 = u_ini.^2;
for i = 1:n
    % Compute log(f(x))
    uL = u_ini(i,:);
    D_gi = zeros(1,p);
    H_gi = eye(p);
    for steps = 1:5
        uL = uL - D_gi/H_gi;
        lmbi = exp(B0*c0+B0*C*uL');
        D_gi = -lmbi'*Phi*dt + sumB(i,:)*C - uL./s2';
        H_gi = -Phi'*diag(lmbi)*Phi*dt - diag(1./s2);
    end
    gi = -sum(lmbi)*dt + sumB(i,:)*(c0+C*uL') - gammaln(m(i)+1) ...
        -sum(uL.^2./(2*s2')) - .5*sum(log(2*pi*s2));
    logf(i) = gi + (p/2)*log(2*pi) - .5*logdet(-H_gi);
    S = (-H_gi)\eye(p);
    u(i,:) = uL;
    u2(i,:) = diag(S)' + uL.^2;
end
end

function y = logdet(A)
% log(det(A)) for symmetric non-neg A
R = chol(A);
y = 2*sum(log(diag(R)));
end

